require 'confstruct'
module Kinetic

  # Configuration can be set either by passing the {Kinetic::DSL#set set} method and or by using the
  # {Kinetic::DSL#config_file config_file} method.
  #
  # @example
  #  require 'kinetic'
  #
  #  config_file 'config.yml' # loads the values in config_file into configuration
  #  set :host,  'localhost'  # sets config.host to 'localhost'
  #
  # Accessing configuration values by their method accessors will raise {Kinetic::Errors::MissingConfigurationValue}
  # error. This can be used to ensure that specific values are set at runtime. To access a value without raising an
  # exception use the hash key instead.
  #
  # @example
  #  config.host   # raises an error if the config host value is not set
  #  config[:host] # returns nil if the host value is not set
  class Configuration < Confstruct::Configuration

    # Default configuration values.
    DEFAULTS = {
      name:       File.basename($0),         # name of the application. This also translates to the name of the AMQP exchange
      root:       File.dirname($0),          # the application root directory
      app_file:   $0,                        # the application file
      workers:    1,                         # the number of worker processes to run
      log_file:   'kinetic.log',             # the path of the log file to be used
      host:       'localhost',               # the AMQP host
      port:       5672,                      # the AMQP port
      serializer: :JSON,                     # the message serializer to use
      timeout:    1,                         # the time (in seconds) to wait for various operations
      pid:        "#{File.basename($0)}.pid" # the process pid file
    }

    def initialize
      super
      self.deep_merge!(DEFAULTS)
    end


    private

    def method_missing(meth, *args, &block)
      value = super
      raise Kinetic::Errors::MissingConfigurationValue.new("Configuration is missing value for #{meth}") unless value
      value
    end

  end
end