module Kinetic
  class Worker

    delegate :logger, :config, :set, :deserialize, to: :app

    attr_reader :id, :app, :channel, :exchanges

    def initialize(id, app)
      @app = app
      logger.debug "Initializing worker #{id}"
      @id = id
      @to_io, @master = Kgio::Pipe.new.each { |io| io.fcntl(Fcntl::F_SETFD, Fcntl::FD_CLOEXEC) }
    end

    def atfork_parent
      logger.debug "Worker #{id} closing pipe to child"
      @to_io = @to_io.close
    end

    def atfork_child
      logger.debug "Worker #{id} closing pipe to master"
      @master = @master.close
    end

    def run
      logger.debug "Establishing connection host: '#{config.host}', port: '#{config.port}'"
      AMQP.start(host: config.host, port: config.port) do |connection|
        logger.debug "AMQP started with conneciton #{connection}"
        initialize_channel!(connection)
        initialize_exchanges!
        initialize_subscribers!
        initialize_timed_events!
      end
    end

    # master sends fake signals to children
    def soft_kill(sig) # :nodoc:
      case sig
      when Integer
        signum = sig
      else
        signum = Signal.list[sig.to_s] or
          raise ArgumentError, "BUG: bad signal: #{sig.inspect}"
      end
      # writing and reading 4 bytes on a pipe is atomic on all POSIX platforms
      # Do not care in the odd case the buffer is full, here.
      @master.kgio_trywrite([signum].pack('l'))
    rescue Errno::EPIPE
      # worker will be reaped soon
    end


    private

    def initialize_channel!(connection)
      logger.debug 'Intitializing channel'
      @channel = AMQP::Channel.new(connection)
    end

    def initialize_exchanges!
      logger.debug 'Initializing exchanges'
      @exchanges = {}
      prefix = config.name
      app.send(:exchanges).each_key do |name|
        logger.debug "   Initializing #{name} exchange"
        @exchanges[name] = channel.send(name, "#{prefix}.#{name}")
      end
    end

    def initialize_subscribers!
      logger.debug 'Initializing subscribers'
      app.send(:exchanges).each do |name, value|
        logger.debug "   Building subscribers for '#{name}' exchange"
        value.each do |queue, block|
          logger.debug "   Initializing queue for '#{queue}'"
          q = channel.queue(queue)
          logger.debug "   Binding queue to '#{name}' exchange with routing_key '#{queue}'"
          q.bind(exchanges[name], routing_key: queue)
          logger.debug "   Subscribing to messages for '#{queue}'"
          q.subscribe do |meta, payload|
            begin
              block.call(deserialize(payload))
            rescue Exception => e
              logger.error e
            end
          end
        end
      end
    end

    def initialize_timed_events!
      logger.debug 'Initializing timed events'
      app.send(:timed_events).each do |label, event|
        logger.debug "   Creating timer for #{label}"
        event[:timer] = EM.add_periodic_timer(event[:interval]) { event[:block].call }
      end
    end

  end
end